/*
    HOTKEYS - use keys on your multimedia keyboard to control your computer
    Copyright (C) 2000  Anthony Y P Wong <ypwong@ypwong.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    $Id: hotkeys.c,v 1.1 2001/01/17 17:14:33 hajime Exp $
*/

#include <X11/Xosdefs.h>
#ifndef X_NOT_STDC_ENV
#include <stdlib.h>
#else
extern char *getenv();
#endif

#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#endif /* HAVE_GETOPT_LONG */

#include <sys/types.h>
#include <sys/stat.h>

/* Mixer related */
#include <fcntl.h>
#include <sys/ioctl.h>
#if defined (__FreeBSD__)
#include <machine/soundcard.h>
#else
#       if defined (__NetBSD__) || defined (__OpenBSD__)
#       include <soundcard.h>          /* OSS emulation */
#       undef ioctl
#       else
/* BSDI, Linux, Solaris */
#       include <sys/soundcard.h>
#       endif                          /* __NetBSD__ or __OpenBSD__ */
#endif                          /* __FreeBSD__ */

/* CDROM related */
#include <linux/cdrom.h>        /* FIXME: linux specific! */
/* APM (suspend/standby) support */
#include "apm.h"

#include "hotkeys.h"


#define	lowbit(x)	((x) & (-(x)))
#define	M(m)	fprintf(stderr,(m))
#define	M1(m,a)	fprintf(stderr,(m),(a))

/***====================================================================***/

char *		dpyName =	NULL;
Display *	dpy =		NULL;
char *		cfgFileName =	NULL;
int		xkbOpcode =	0;
int		xkbEventCode =	0;

unsigned long	eventMask =	0;

Bool		synch =		False;
int		verbose =	0;
Bool		background =	False;

char *		soundCmd =	NULL;
char *		soundDir =	NULL;

XkbDescPtr	xkb =		NULL;

FILE *          errorFile =     NULL;

char *          progname =      NULL;

char *          cdromDevice =   CDROM_DEV;

keyboard        kbd;                    /* the keyboard the user is using */

/***====================================================================***/

static Display *
GetDisplay(char* program, char* dpyName, int* opcodeRtrn, int* evBaseRtrn)
{
    int	mjr,mnr,error;
    Display	*dpy;

    mjr= XkbMajorVersion;
    mnr= XkbMinorVersion;
    dpy= XkbOpenDisplay(dpyName,evBaseRtrn,NULL,&mjr,&mnr,&error);
    if (dpy==NULL) {
	switch (error) {
	    case XkbOD_BadLibraryVersion:
		uInfo("%s was compiled with XKB version %d.%02d\n",
				program,XkbMajorVersion,XkbMinorVersion);
		uInfo("X library supports incompatible version %d.%02d\n",
				mjr,mnr);
		break;
	    case XkbOD_ConnectionRefused:
		uInfo("Cannot open display \"%s\"\n",dpyName);
		break;
	    case XkbOD_NonXkbServer:
		uInfo("XKB extension not present on %s\n",dpyName);
		break;
	    case XkbOD_BadServerVersion:
		uInfo("%s was compiled with XKB version %d.%02d\n",
				program,XkbMajorVersion,XkbMinorVersion);
		uInfo("Server %s uses incompatible version %d.%02d\n",
				dpyName,mjr,mnr);
		break;
	    default:
		uInternalError("Unknown error %d from XkbOpenDisplay\n",error);
	}
    }
    else
    {
        if (synch)
            XSynchronize(dpy,True);
        if (opcodeRtrn)
            XkbQueryExtension(dpy,opcodeRtrn,evBaseRtrn,NULL,&mjr,&mnr);
    }
    return dpy;
}

void
bailout(void)
{
//    uAction("Exiting\n");
    if ( dpy != NULL )
	XCloseDisplay(dpy);
    uInfo("Bailing out...\n");
    exit(1);
}

/***====================================================================***/

static void
printXkbActionMessage(FILE* file,XkbEvent* xkbev)
{
    XkbActionMessageEvent *msg= &xkbev->message;
    fprintf(file,"    message: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
                                        msg->message[0],msg->message[1],
                                        msg->message[2],msg->message[3],
                                        msg->message[4],msg->message[5]);
    fprintf(file,"    key %d, event: %s,  follows: %s\n",msg->keycode,
                                     (msg->press?"press":"release"),
                                     (msg->key_event_follows?"yes":"no"));
    return;
}


void
uError(char* s,...)
{
    va_list ap;

    va_start(ap, s);
    fprintf(errorFile,"%s: ", progname);
    vfprintf(errorFile,s,ap);
    fprintf(errorFile,": %s\n", strerror(errno));
    fflush(errorFile);
    va_end(ap);
    return;
}

void
uInfo(char* s,...)
{
    va_list ap;

    va_start(ap, s);
    fprintf(errorFile,"%s: ", progname);
    vfprintf(errorFile,s,ap);
    fflush(errorFile);
    va_end(ap);
    return;
}

void
uInternalError(char* s,...)
{
    va_list ap;

    va_start(ap, s);
    fprintf(errorFile,"%s: internal error: ", progname);
    vfprintf(errorFile,s,ap);
    fflush(errorFile);
    va_end(ap);
    return;
}

/***====================================================================***/

static void
initialize(char* argv[])
{
    KeySym              newKS;
    XkbMessageAction    xma;
    XkbMapChangesRec    mapChangeRec;
    int                 types[1];
    int                 i;
    int                 code = 250;
int j=0;

    dpy = GetDisplay(argv[0], dpyName, &xkbOpcode, &xkbEventCode);
    if (!dpy)
        bailout();

    /* Select the ActionMessage event */
    if ( !XkbSelectEvents( dpy, XkbUseCoreKbd, XkbActionMessageMask,
                           XkbActionMessageMask ))
    {
        uInfo("Couldn't select desired XKB events\n");
        bailout();
    }
//    xkb= XkbGetKeyboard(dpy,XkbGBN_AllComponentsMask,XkbUseCoreKbd);
//    xkb= XkbGetKeyboard(dpy,XkbAllComponentsMask,XkbUseCoreKbd);
    xkb = XkbGetMap(dpy, XkbAllMapComponentsMask, XkbUseCoreKbd);
    if (!xkb)
    {
        uInfo("XkbGetMap failed\n"); bailout();
    }

    /* Construct the Message Action struct */
    xma.type = XkbSA_ActionMessage;
    xma.flags = XkbSA_MessageOnPress;
    strcpy(xma.message," ");

#ifdef DEBUG
printf("num_acts:%d size_acts:%d\n", xkb->server->num_acts, xkb->server->size_acts);
printf("idx:%d has action:%d no.:%d noOfGrps:%d\n", xkb->server->key_acts[152], XkbKeyHasActions(xkb,152), XkbKeyNumActions(xkb,152), XkbKeyNumGroups(xkb,152));
#endif

        /* Assign a group to the key code */
//        types[XkbGroup1Index] = XkbKeyTypeIndex( xkb, code, XkbGroup1Index );
        types[0] = XkbOneLevelIndex;
        XkbChangeTypesOfKey(xkb, code, 1, XkbGroup1Mask, types, NULL);

        /* Change their Keysyms */
        if ( XkbResizeKeySyms( xkb, code, 1) == NULL )
        {
            uInfo("resize keysym failed\n"); bailout();
        }
        /* Add one key action to it */
        if ( XkbResizeKeyActions( xkb, code, 1) == NULL )
        {
            uInfo("resize key action failed\n"); bailout();
        }

        /* Assign a new keysym to the key code */
        newKS = 0x2200+code;       /* FIXME: I just choose a keysym that's not yet assigned */
//        newKS = 0xFFE0;
        *XkbKeySymsPtr(xkb,code) = newKS;

        /* Send the change back to the server */
        /* XkbKeyActionsMask must be here, just XkbKeySymsMask|XkbKeyTypesMask
         * is not sufficient */
#if 0
        if ( XkbSetMap(dpy, XkbKeySymsMask|XkbKeyActionsMask|XkbKeyTypesMask, xkb) )
        {
#ifdef DEBUG
            printf("map set done\n");
#endif
        }
        else
        {
            uInfo("map set failed\n"); bailout();
        }
#endif

hahaha:
#ifdef DEBUG
        printf("%d: before: %d\n",code, XkbKeyActionsPtr(xkb,code)[0].type);
#endif
        /* Assign the Message Action to the key code */
        (&(xkb->server->acts[ xkb->server->key_acts[code] ]))[0] = (XkbAction) xma;

#ifdef DEBUG
        printf("after: %d\n",XkbKeyActionsPtr(xkb,code)[0].type);
#endif

    /***************************************************************/

#if 0
    /* Send the change back to the server */
    if ( XkbSetMap(dpy, XkbAllMapComponentsMask, xkb) )
    {
#ifdef DEBUG
        printf("map set done\n");
#endif
    }
    else
    {
        uError("map set failed\n"); bailout();
    }
#endif

    /* Commit the change back to the server */
    bzero(&mapChangeRec, sizeof(mapChangeRec));
    mapChangeRec.changed = XkbKeySymsMask | XkbKeyTypesMask;
    mapChangeRec.first_key_sym = code; /* FIXME, the value is from experimentation */
    mapChangeRec.num_key_syms = 1;  /* FIXME, the value is from experimentation  */
    mapChangeRec.first_key_act = code; /* FIXME, the value is from experimentation */
    mapChangeRec.num_key_acts = 1;  /* FIXME, the value is from experimentation  */
    mapChangeRec.first_type = 0;
    mapChangeRec.num_types = xkb->map->num_types;
    if ( XkbChangeMap(dpy, xkb, &mapChangeRec) )
    {
#ifdef DEBUG
        printf("map changed done\n");
#endif
    }
    else
    {
        uError("map changed failed\n"); bailout();
    }
        (&(xkb->server->acts[ xkb->server->key_acts[code] ]))[0] = (XkbAction) xma;

#if 0
    /* Commit the change back to the server. This is necessary! */
    if ( XkbSetMap(dpy, XkbKeyActionsMask, xkb) )
    {
#ifdef DEBUG
        printf("map set done\n");
#endif
    }
    else
    {
        uInfo("map set failed\n"); bailout();
    }
#endif
    /* Commit the change back to the server */
    bzero(&mapChangeRec, sizeof(mapChangeRec));
    mapChangeRec.changed = XkbKeyActionsMask;
    mapChangeRec.first_key_act = code; /* FIXME, the value is from experimentation */
    mapChangeRec.num_key_acts = 1;  /* FIXME, the value is from experimentation  */
    if ( XkbChangeMap(dpy, xkb, &mapChangeRec) )
    {
#ifdef DEBUG
        printf("map changed done\n");
#endif
    }
    else
    {
        uError("map changed failed\n"); bailout();
    }

/* _test_ */
if ( xkb=XkbGetMap(dpy,XkbAllMapComponentsMask,XkbUseCoreKbd) )
    printf("map get done\n");
else {
    printf("map get failed\n"); exit(0);
}

printf("after: %d\n",XkbKeyActionsPtr(xkb,152)[0].type);
printf("idx:%d has action:%d no.:%d noOfGrps:%d\n", xkb->server->key_acts[38], XkbKeyHasActions(xkb,38), XkbKeyNumActions(xkb,38), XkbKeyNumGroups(xkb,38));
printf("idx:%d has action:%d no.:%d noOfGrps:%d\n", xkb->server->key_acts[152], XkbKeyHasActions(xkb,152), XkbKeyNumActions(xkb,152), XkbKeyNumGroups(xkb,152));

printf("num_acts:%d size_acts:%d\n", xkb->server->num_acts, xkb->server->size_acts);
/*
XkbSelectEvents(dpy,XkbUseCoreKbd,XkbAllEventsMask,XkbAllEventsMask);
*/
}


int
main(int argc, char *argv[])
{

    initialize(argv);
}

